@@ -2,11 +2,10 @@
 
 use crate::read::magic_finder::{Backwards, Forward, MagicFinder, OptimisticMagicFinder};
 use crate::read::ArchiveOffset;
-use crate::result::{ZipError, ZipResult};
+use crate::result::{invalid, ZipError, ZipResult};
 use core::mem;
-use std::io;
-use std::io::prelude::*;
-use std::slice;
+use core::slice;
+use std::io::{self, Read, Seek, Write};
 
 /// "Magic" header values used in the zip spec to locate metadata records.
 ///
@@ -49,6 +48,7 @@ impl Magic {
     pub const CENTRAL_DIRECTORY_END_SIGNATURE: Self = Self::literal(0x06054b50);
     pub const ZIP64_CENTRAL_DIRECTORY_END_SIGNATURE: Self = Self::literal(0x06064b50);
     pub const ZIP64_CENTRAL_DIRECTORY_END_LOCATOR_SIGNATURE: Self = Self::literal(0x07064b50);
+    pub const DATA_DESCRIPTOR_SIGNATURE: Self = Self::literal(0x08074b50);
 }
 
 /// Similar to [`Magic`], but used for extra field tags as per section 4.5.3 of APPNOTE.TXT.
@@ -91,7 +91,7 @@ impl ExtraFieldMagic {
 /// The file size at which a ZIP64 record becomes necessary.
 ///
 /// If a file larger than this threshold attempts to be written, compressed or uncompressed, and
-/// [`FileOptions::large_file()`](crate::write::FileOptions) was not true, then [`ZipWriter`] will
+/// [`FileOptions::large_file()`](crate::write::FileOptions::large_file) was not true, then [`crate::ZipWriter`] will
 /// raise an [`io::Error`] with [`io::ErrorKind::Other`].
 ///
 /// If the zip file itself is larger than this value, then a zip64 central directory record will be
@@ -101,33 +101,38 @@ impl ExtraFieldMagic {
 /// # fn main() -> Result<(), zip::result::ZipError> {
 /// # #[cfg(target_pointer_width = "64")]
 /// # {
-/// use std::io::{self, Cursor, prelude::*};
+/// use std::io::{self, Cursor, Write};
 /// use std::error::Error;
 /// use zip::{ZipWriter, write::SimpleFileOptions};
 ///
 /// let mut zip = ZipWriter::new(Cursor::new(Vec::new()));
 /// // Writing an extremely large file for this test is faster without compression.
-/// let options = SimpleFileOptions::default().compression_method(zip::CompressionMethod::Stored);
 ///
 /// let big_len: usize = (zip::ZIP64_BYTES_THR as usize) + 1;
 /// let big_buf = vec![0u8; big_len];
-/// zip.start_file("zero.dat", options)?;
-/// // This is too big!
-/// let res = zip.write_all(&big_buf[..]).err().unwrap();
-/// assert_eq!(res.kind(), io::ErrorKind::Other);
-/// let description = format!("{}", &res);
-/// assert_eq!(description, "Large file option has not been set");
-/// // Attempting to write anything further to the same zip will still succeed, but the previous
-/// // failing entry has been removed.
-/// zip.start_file("one.dat", options)?;
-/// let zip = zip.finish_into_readable()?;
-/// let names: Vec<_> = zip.file_names().collect();
-/// assert_eq!(&names, &["one.dat"]);
+/// {
+///     let options = SimpleFileOptions::default()
+///         .compression_method(zip::CompressionMethod::Stored);
+///     zip.start_file("zero.dat", options)?;
+///     // This is too big!
+///     let res = zip.write_all(&big_buf[..]).err().unwrap();
+///     assert_eq!(res.kind(), io::ErrorKind::Other);
+///     let description = format!("{}", &res);
+///     assert_eq!(description, "Large file option has not been set");
+///     // Attempting to write anything further to the same zip will still succeed, but the previous
+///     // failing entry has been removed.
+///     zip.start_file("one.dat", options)?;
+///     let zip = zip.finish_into_readable()?;
+///     let names: Vec<_> = zip.file_names().collect();
+///     assert_eq!(&names, &["one.dat"]);
+/// }
 ///
 /// // Create a new zip output.
 /// let mut zip = ZipWriter::new(Cursor::new(Vec::new()));
 /// // This time, create a zip64 record for the file.
-/// let options = options.large_file(true);
+/// let options = SimpleFileOptions::default()
+///      .compression_method(zip::CompressionMethod::Stored)
+///      .large_file(true);
 /// zip.start_file("zero.dat", options)?;
 /// // This succeeds because we specified that it could be a large file.
 /// assert!(zip.write_all(&big_buf[..]).is_ok());
@@ -139,7 +144,7 @@ pub const ZIP64_BYTES_THR: u64 = u32::MAX as u64;
 /// The number of entries within a single zip necessary to allocate a zip64 central
 /// directory record.
 ///
-/// If more than this number of entries is written to a [`ZipWriter`], then [`ZipWriter::finish()`]
+/// If more than this number of entries is written to a [`crate::ZipWriter`], then [`crate::ZipWriter::finish()`]
 /// will write out extra zip64 data to the end of the zip file.
 pub const ZIP64_ENTRY_THR: usize = u16::MAX as usize;
 
@@ -266,8 +271,7 @@ impl FixedSizeBlock for Zip32CDEBlock {
         self.magic
     }
 
-    const WRONG_MAGIC_ERROR: ZipError =
-        ZipError::InvalidArchive("Invalid digital signature header");
+    const WRONG_MAGIC_ERROR: ZipError = invalid!("Invalid digital signature header");
 
     to_and_from_le![
         (magic, Magic),
@@ -333,9 +337,7 @@ impl Zip32CentralDirectoryEnd {
         let mut zip_file_comment = vec![0u8; zip_file_comment_length as usize].into_boxed_slice();
         if let Err(e) = reader.read_exact(&mut zip_file_comment) {
             if e.kind() == io::ErrorKind::UnexpectedEof {
-                return Err(ZipError::InvalidArchive(
-                    "EOCD comment exceeds file boundary",
-                ));
+                return Err(invalid!("EOCD comment exceeds file boundary"));
             }
 
             return Err(e.into());
@@ -356,9 +358,7 @@ impl Zip32CentralDirectoryEnd {
         let (block, comment) = self.into_block_and_comment();
 
         if comment.len() > u16::MAX as usize {
-            return Err(ZipError::InvalidArchive(
-                "EOCD comment length exceeds u16::MAX",
-            ));
+            return Err(invalid!("EOCD comment length exceeds u16::MAX"));
         }
 
         block.write(writer)?;
@@ -390,8 +390,7 @@ impl FixedSizeBlock for Zip64CDELocatorBlock {
         self.magic
     }
 
-    const WRONG_MAGIC_ERROR: ZipError =
-        ZipError::InvalidArchive("Invalid zip64 locator digital signature header");
+    const WRONG_MAGIC_ERROR: ZipError = invalid!("Invalid zip64 locator digital signature header");
 
     to_and_from_le![
         (magic, Magic),
@@ -467,8 +466,7 @@ impl FixedSizeBlock for Zip64CDEBlock {
         self.magic
     }
 
-    const WRONG_MAGIC_ERROR: ZipError =
-        ZipError::InvalidArchive("Invalid digital signature header");
+    const WRONG_MAGIC_ERROR: ZipError = invalid!("Invalid digital signature header");
 
     to_and_from_le![
         (magic, Magic),
@@ -513,11 +511,9 @@ impl Zip64CentralDirectoryEnd {
         } = Zip64CDEBlock::parse(reader)?;
 
         if record_size < 44 {
-            return Err(ZipError::InvalidArchive("Low EOCD64 record size"));
+            return Err(invalid!("Low EOCD64 record size"));
         } else if record_size.saturating_add(12) > max_size {
-            return Err(ZipError::InvalidArchive(
-                "EOCD64 extends beyond EOCD64 locator",
-            ));
+            return Err(invalid!("EOCD64 extends beyond EOCD64 locator"));
         }
 
         let mut zip_file_comment = vec![0u8; record_size as usize - 44].into_boxed_slice();
@@ -639,7 +635,7 @@ pub(crate) fn find_central_directory<R: Read + Seek>(
         // ! Relaxed (inequality) due to garbage-after-comment Python files
         // Consistency check: the EOCD comment must terminate before the end of file
         if eocd.zip_file_comment.len() as u64 + eocd_offset + 22 > file_len {
-            parsing_error = Some(ZipError::InvalidArchive("Invalid EOCD comment length"));
+            parsing_error = Some(invalid!("Invalid EOCD comment length"));
             continue;
         }
 
@@ -649,9 +645,7 @@ pub(crate) fn find_central_directory<R: Read + Seek>(
                 eocd_offset: u64,
             ) -> ZipResult<(u64, Zip64CentralDirectoryEndLocator)> {
                 if eocd_offset < mem::size_of::<Zip64CDELocatorBlock>() as u64 {
-                    return Err(ZipError::InvalidArchive(
-                        "EOCD64 Locator does not fit in file",
-                    ));
+                    return Err(invalid!("EOCD64 Locator does not fit in file"));
                 }
 
                 let locator64_offset = eocd_offset - mem::size_of::<Zip64CDELocatorBlock>() as u64;
@@ -683,7 +677,7 @@ pub(crate) fn find_central_directory<R: Read + Seek>(
 
             // Consistency check: the CD relative offset cannot be after the EOCD
             if relative_cd_offset >= eocd_offset {
-                parsing_error = Some(ZipError::InvalidArchive("Invalid CDFH offset in EOCD"));
+                parsing_error = Some(invalid!("Invalid CDFH offset in EOCD"));
                 continue;
             }
 
@@ -715,20 +709,18 @@ pub(crate) fn find_central_directory<R: Read + Seek>(
                 });
             }
 
-            parsing_error = Some(ZipError::InvalidArchive("No CDFH found"));
+            parsing_error = Some(invalid!("No CDFH found"));
             continue;
         };
 
         // Consistency check: the EOCD64 offset must be before EOCD64 Locator offset */
         if locator64.end_of_central_directory_offset >= locator64_offset {
-            parsing_error = Some(ZipError::InvalidArchive("Invalid EOCD64 Locator CD offset"));
+            parsing_error = Some(invalid!("Invalid EOCD64 Locator CD offset"));
             continue;
         }
 
         if locator64.number_of_disks > 1 {
-            parsing_error = Some(ZipError::InvalidArchive(
-                "Multi-disk ZIP files are not supported",
-            ));
+            parsing_error = Some(invalid!("Multi-disk ZIP files are not supported"));
             continue;
         }
 
@@ -743,16 +735,12 @@ pub(crate) fn find_central_directory<R: Read + Seek>(
 
             // Consistency check: EOCD64 locator should agree with the EOCD64
             if z64.disk_with_central_directory != locator64.disk_with_central_directory {
-                return Err(ZipError::InvalidArchive(
-                    "Invalid EOCD64: inconsistency with Locator data",
-                ));
+                return Err(invalid!("Invalid EOCD64: inconsistency with Locator data"));
             }
 
             // Consistency check: the EOCD64 must have the expected length
             if z64.record_size + 12 != expected_length {
-                return Err(ZipError::InvalidArchive(
-                    "Invalid EOCD64: inconsistent length",
-                ));
+                return Err(invalid!("Invalid EOCD64: inconsistent length"));
             }
 
             Ok(z64)
@@ -795,9 +783,8 @@ pub(crate) fn find_central_directory<R: Read + Seek>(
                             )
                             .saturating_add(eocd64.central_directory_offset)
                     {
-                        local_error = Some(ZipError::InvalidArchive(
-                            "Invalid EOCD64: inconsistent number of files",
-                        ));
+                        local_error =
+                            Some(invalid!("Invalid EOCD64: inconsistent number of files"));
                         continue;
                     }
 
@@ -813,10 +800,10 @@ pub(crate) fn find_central_directory<R: Read + Seek>(
             }
         }
 
-        parsing_error = local_error.or(Some(ZipError::InvalidArchive("Could not find EOCD64")));
+        parsing_error = local_error.or(Some(invalid!("Could not find EOCD64")));
     }
 
-    Err(parsing_error.unwrap_or(ZipError::InvalidArchive("Could not find EOCD")))
+    Err(parsing_error.unwrap_or(invalid!("Could not find EOCD")))
 }
 
 pub(crate) fn is_dir(filename: &str) -> bool {
@@ -828,9 +815,13 @@ pub(crate) fn is_dir(filename: &str) -> bool {
 
 #[cfg(test)]
 mod test {
-    use super::*;
     use std::io::Cursor;
 
+    use crate::{
+        result::{invalid, ZipError},
+        spec::{FixedSizeBlock, Magic, Pod},
+    };
+
     #[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
     #[repr(packed, C)]
     pub struct TestBlock {
@@ -847,7 +838,7 @@ mod test {
             self.magic
         }
 
-        const WRONG_MAGIC_ERROR: ZipError = ZipError::InvalidArchive("unreachable");
+        const WRONG_MAGIC_ERROR: ZipError = invalid!("unreachable");
 
         to_and_from_le![(magic, Magic), (file_name_length, u16)];
     }
